import os

import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')

import numpy as np
import torch

import torch.nn as nn
import anyact as aa

from gradiator import gradiator

import h5py

import torchvision
import SOLO_shared
mseloss = nn.MSELoss()
cosloss = aa.CosLoss()
dotloss = aa.DotLoss()
dotsim = aa.DotSim()

class SOLO_bwd(nn.Module):
	def __init__(self):
		super().__init__()
		# c1=[1,32,32,32,32,32,32,32,32,64]
		c1 = [1,32,64]
		cf = c1[-1]
		self.layer_x = aa.make_nn_cnn_real(c1,ker_size=5,pad=2,σs=[nn.Identity(),nn.GELU()])
		self.layer_g = aa.make_nn_cnn_real(c1,ker_size=5,pad=2,σs=[nn.Identity(),nn.GELU()])
		self.layer_xg = nn.Sequential(
			### Version A
			# aa.ToComplex(),aa.Log(),
			# aa.make_nn_cnn_real([2*cf,cf,cf//4],ker_size=5,pad=0,σs=[aa.GELU()]),
			# aa.Exp(),aa.ToReal(),
			### Version B
			aa.make_nn_cnn_real([2*cf,cf,cf//4],ker_size=5,pad=0,σs=[nn.GELU()]),
			###
			nn.ConvTranspose2d(cf//4,cf//16,2,stride=2),
			nn.ConvTranspose2d(cf//16,cf//64,2,stride=2),
		)
		self.crop = torchvision.transforms.CenterCrop((64,64))
	def forward(self,x,g):
		xg = torch.cat((self.layer_x(x),self.layer_g(g)),1)
		xg = nn.MaxPool2d(2)(xg)
		xg = self.layer_xg(xg)
		# print(xg.shape)
		xg = self.crop(xg)
		return xg

def debug_bwd():
	x = torch.randn(10,1,64,64)
	g = torch.randn(10,1,64,64)
	bwd = SOLO_bwd()
	print(bwd(x,g).shape == x.shape)
	return bwd(x,g).shape

# debug_bwd()

class SOLO_derivable(nn.Module):
	def to(self,device):
		print(f"to device {device}")
		nn.Module.to(self,device)
		self.device=device
		self.hidden["bwd_model"] = self.hidden["bwd_model"].to(device)
		return self
	def __init__(self,opti_mode,bwd_lr,loss_mode,identifier,bwd_sampling=2**-20,stupid_mode=False):
		super(SOLO_derivable,self).__init__()
		self.identifier = identifier
		self.device='cpu'
		self.hidden = {}
		self.loss_mode=loss_mode
		def fwd(x):
			x = SOLO_shared.SOLO_scale(x,margin=[0,0,0,0],res=[128,128],mode="bilinear")
			x = SOLO_shared.SOLO_device(x, stupid_mode=stupid_mode)
			x = SOLO_shared.SOLO_scale(x,margin=[0,0,0,0],res=[64,64],mode="bilinear")
			return x
		self.fwd = fwd
		self.hidden["bwd_model"] = SOLO_bwd()
		self.hist_x=[]
		self.hist_y=[]
		self.hist_g=[]
		def bwd(g,y,x,save=True):
			# print("bwd !")
			torch.cuda.empty_cache() #########################################################################################################################
			if save:
				self.hist_x.append(x.detach().cpu())
				self.hist_y.append(y.detach().cpu())
				self.hist_g.append(g.detach().cpu())
				self.learn()
			return self.hidden["bwd_model"](x,g)
		self.bwd = bwd
		### BACKWARD TRAINING
		self.bwd_epoch = 0
		self.bwd_epoch_rel = 0
		dim_inputs = 128*128; dim_outputs = 128*128
		self.bwd_dim = dim_inputs*dim_outputs
		self.memory_limit = 8192 # we cannot store the full history
		self.bwd_sampling = bwd_sampling
		if   opti_mode=="adam": self.bwd_optim = torch.optim.Adam(self.hidden["bwd_model"].parameters(),lr=bwd_lr,eps=1e-3)
		elif opti_mode=="sgd" : self.bwd_optim = torch.optim.SGD (self.hidden["bwd_model"].parameters(),lr=bwd_lr)
		else : print("error"); return
		self.hist = {
			"loss_mse" : [],
			"loss_cos" : [],
			"loss_dot" : [],
			"loss_dotsim" : [],
			"epochs_rel": [],
		}
		### FINALIZE SETUP
		self.gradiator = gradiator(self.fwd,self.bwd)
	def learn(self):
		print("blaxkbox bwd learn !")
		########## learn a little more
		# forget older entries, target data worth dim_i*dim_o epochs (ie 1 epoch_rel)
		population_per_sample = np.mean([s.shape[0] for s in self.hist_x])
		memory_limit = int(.5+self.memory_limit/population_per_sample)
		# memory_limit = 128
		# print(f"I have {np.mean([s.shape[0] for s in self.hist_x])}")
		self.hist_x = self.hist_x[-memory_limit:]
		self.hist_y = self.hist_y[-memory_limit:]
		self.hist_g = self.hist_g[-memory_limit:]
		sample_count=0
		NNN = int(.5 + self.bwd_dim*self.bwd_sampling)
		shuffle=False # FIXME : It's buggy !!!!! we should not mix each elements x,y,go !
		# for learn_step,(x,y,go) in enumerate(zip(self.hist_x,self.hist_y,self.hist_g)):
		for learn_step,(x,y,go) in enumerate(zip(
			SOLO_shared.HistIterator(self.hist_x,NNN,shuffle),
			SOLO_shared.HistIterator(self.hist_y,NNN,shuffle),
			SOLO_shared.HistIterator(self.hist_g,NNN,shuffle),
		)):
			torch.cuda.empty_cache() #########################################################################################################################
			self.bwd_epoch += 1
			self.bwd_epoch_rel += x.shape[0]/self.bwd_dim
			sample_count += x.shape[0]
			# print(f"learning step {learn_step}, h={len(self.hist_x)}")
			self.hidden["bwd_model"].train()
			x = x.to(self.device)
			y = y.to(self.device)
			go=go.to(self.device)
			ε = 1e-2 # amplitude of dx ; TODO : something with optimizer ?
			# on peut prendre ε=0.01 pour être de l'ordre de 2/255
			dx = (torch.randn(x.shape)*ε).to(self.device)                    #   x+ε  -   x  =  ε
			df = self.fwd(x+dx) - y # with y=self(x)          # A(x+ε) - A(x) = Aε
			gx = self.bwd(x,y,go,False)
			should = (gx*dx).sum(dim=1).sum(dim=1).sum(dim=1).view(1,-1) # for subspace
			found  = (go*df).sum(dim=1).sum(dim=1).sum(dim=1).view(1,-1) # for subspace (corresponding ones)
			# sample_dim = NNN
			# dx = ε*torch.randn(*x.shape,sample_dim).to(self.device)
			# df = self.fwd(x+dx) - y # y=f(x)
			# gx = self.bwd(x,y,go,False)
			# to_sum = list(range(1,len(x.shape)))
			# go = go.unsqueeze(len(go.shape)).expand(*go.shape,sample_dim)
			# should = (gx*dx).sum(to_sum)
			# found  = (go*df).sum(to_sum)
			#
			loss_mse = mseloss(should, found)
			loss_cos = cosloss(should, found)
			loss_dot = dotloss(found,should)
			loss_dotsim = 1 - dotsim(found,should).mean()
			    # I expect loss_dotsim to be shit for learning, cf illustration on laurent.sexy/nn/pres/
			    # but we keep it only to check loss progress
			# 
			self.bwd_optim.zero_grad()
			if   self.loss_mode=="mse": loss_mse.requires_grad_().backward()
			elif self.loss_mode=="cos": loss_cos.requires_grad_().backward()
			# elif self.loss_mode=="dot": loss_dotsim.backward()
			elif self.loss_mode=="dot": loss_dot.requires_grad_().backward()
			elif self.loss_mode=="mse+cos": (loss_mse+loss_cos).requires_grad_().backward()
			elif self.loss_mode=="mse+dot": (loss_mse+loss_dot).requires_grad_().backward()
			elif self.loss_mode=="cos+dot": (loss_cos+loss_dot).requires_grad_().backward()
			elif self.loss_mode=="mse,cos":
				rnd = torch.randint(2,(1,))
				if rnd==0 : loss_mse.requires_grad_().backward()
				if rnd==1 : loss_cos.requires_grad_().backward()
				# if rnd==2 : loss_dot.requires_grad_().backward()
			# elif loss_mode=="mse^cos": torch.max(loss_mse,loss_cos).backward()
			# elif loss_mode=="mse^dot": torch.max(loss_mse,loss_dot).backward()
			else : print("error"); return
			self.bwd_optim.step()
			# 
			self.hist["loss_mse"].append(loss_mse.item())
			self.hist["loss_cos"].append(loss_cos.item())
			self.hist["loss_dot"].append(loss_dot.item())
			self.hist["loss_dotsim"].append(loss_dotsim.item())
			self.hist["epochs_rel"].append(self.bwd_epoch_rel)
			# print(f"bwd_epoch {self.bwd_epoch:08d} → {self.bwd_epoch_rel:7.7f} | total_sample_count={sample_count}")
			print(f"bwd_epoch {self.bwd_epoch_rel:7.7f} | current_sample_count={x.shape[0]:7d} | total_sample_count={sample_count:7d} | mse={loss_mse.item():3.5f} | cos={loss_cos.item():3.5f} | dot={loss_dot.item():3.5f}")
		########## save progress
		self.hidden["bwd_model"].eval()
		os.makedirs(f"out/",exist_ok=True)
		es = self.hist["epochs_rel"] # np.linspace(0,self.bwd_epoch_rel,self.bwd_epoch) # TODO COMPUTE relative epoch
		def figure_saver(who,name,color,log=True):
			def nplog(x): return np.log(np.array(x))
			plt.figure(figsize=(10,5));plt.tight_layout(pad=0)
			y = self.hist[who]
			if log: y = nplog(y)
			# plt.plot(es,y,c=color,label=name,linewidth=1)
			plt.scatter(es,y,c=color,label=name,s=1)
			plt.savefig(f'out/{self.identifier}-bwd-{name}.png');plt.close("all")
			# pltpath = f'out/{self.identifier}-bwd-{name}'
			# plt.savefig(f'{pltpath}-new.png');plt.close() # do not overwrite directly
			# os.system(f'mv "{pltpath}-new.png" "{pltpath}.png"')
		# 
		figure_saver("loss_mse","loss-mse","blue")
		figure_saver("loss_cos","loss-cos","red")
		figure_saver("loss_dot","loss-dot","green")
		figure_saver("loss_dotsim","loss-dotsim","green",False)
		with h5py.File(f'out/{self.identifier}-bwd.h5', 'w') as file:
			for k in self.hist: file.create_dataset(k, data=self.hist[k])
	def forward(self,x):
		return self.gradiator(x)

# class SOLO_derivable(nn.Module):
# 	def to(self,device):
# 		print(f"to device {device}")
# 		nn.Module.to(self,device)
# 		self.device=device
# 		self.hidden["bwd_model"] = self.hidden["bwd_model"].to(device)
# 		return self
# 	def __init__(self,opti_mode,bwd_lr,loss_mode,identifier,stupid_mode=False):
# 		super(SOLO_derivable,self).__init__()
# 		self.device='cpu'
# 		self.hidden = {}
# 		def fwd(x): return x # SOLO_shared.SOLO_device(x, stupid_mode=stupid_mode)
# 		self.fwd = fwd
# 		# self.hidden["bwd_model"] = SOLO_bwd()
# 		def bwd(g,y,x,save=True):
# 			return x
# 			# return self.hidden["bwd_model"](x,g)
# 		self.bwd = bwd
# 		self.gradiator = gradiator(self.fwd,self.bwd)
# 	def forward(self,x):
# 		return self.gradiator(x)

def debug_SOLO(device="cpu",stupid_mode=True):
	print("INIT MODEL")
	s = SOLO_derivable("adam",1e-2,"mse+cos","debugging",stupid_mode=stupid_mode).to(device)
	C = nn.Conv2d(1,1,3,padding=1).to(device)
	print("OPTIMIZER SETUP")
	optim = torch.optim.Adam(C.parameters(),lr=1e-3)
	for epoch in range(256):
		N= 32 # ! memory limit... I think we can't do much more than that...
		print("INIT INPUTS")
		x = torch.randn(N,1,64,64).to(device)
		g = torch.randn(N,1,64,64).to(device)
		print("EVALUATION")
		y1 = s(C(x))
		y2 = torch.randn(N,1,64,64).to(device)
		print("LOSS COMPUTE")
		loss = nn.MSELoss()(y1,y2)
		print("OPTIM STEP")
		optim.zero_grad()
		loss.requires_grad_().backward()
		optim.step()

# debug_SOLO("cuda")
# debug_SOLO("cpu")
# debug_SOLO("cpu",False)
